/*****************************************************************************
 *
 * Atmel Corporation
 *
 * ̧		: AVRBootloader.cpp
 * ߲	: Dev-C++ 4.9.8.0 - http://bloodshed.net/dev/
 * 		: $Revision: 1163 $
 * t		: $Date: 2006N8215:38:16 +0200 $
 * XV	: $Author: ohlia $
 *
 * xҰ	: avr@atmel.com
 *
 * ړI	: Win32
 *
 * pLq	: AVR911 - AVRJ݊
 *
 * 		: AVR109pLqŋLqAVRް۰ނɑ΂Ĳ̪񋟂
 *			  ׽B̸׽AVRProgrammerh܂B
 *
 ****************************************************************************/
#include "AVRBootloader.hpp"

#include <sstream>
#include <iomanip>

#define MEM_PROGRESS_GRANULARITY 256	// i\ł̎gp̂


/* ݽ׸ */
AVRBootloader::AVRBootloader( CommChannel * _comm ) :
	AVRProgrammer::AVRProgrammer( _comm )
{
	/* ̺ނȂ */
}


/* ޽׸ */
AVRBootloader::~AVRBootloader()
{
	/* ̺ނȂ */
}


bool AVRBootloader::enterProgrammingMode()
{
	return true;							// ް۰ނɑ΂ďOK
}


bool AVRBootloader::leaveProgrammingMode()
{
	return true;							// ް۰ނɑ΂ďOK
}


bool AVRBootloader::chipErase()
{
	/* 'e'ߑM */
	comm->sendByte( 'e' );
	comm->flushTX();

	/* CRԂׂ */
	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Chip erase failed! "
							"Programmer did not return CR after 'e'-command." );

	return true;							// x閽߂B
}


bool AVRBootloader::readOSCCAL( long pos, long * value )
{
	return false;							// xȂ߂B
}

unsigned char AVRBootloader::RC_Calibrate()
{
     return false;
}

bool AVRBootloader::readSignature( long * sig0, long * sig1, long * sig2 )
{
	/* 's'ߑM */
	comm->sendByte( 's' );
	comm->flushTX();

	/* ۂ̎[擾 */
	*sig2 = comm->getByte();
	*sig1 = comm->getByte();
	*sig0 = comm->getByte();
}


bool AVRBootloader::checkSignature( long sig0, long sig1, long sig2 )
{
	long sig[3];

	/* [擾 */
	readSignature( sig, sig+1, sig+2 );

	/* [r */
	if( sig[0] != sig0 || sig[1] != sig1 || sig[2] != sig2 )
	{
		ostringstream msg;
		msg << "Signature does not match selected device! ";
		msg << "Actual signature: (" << hex
			<< "0x" << setw(2) << sig[0] << " "
			<< "0x" << setw(2) << sig[1] << " "
			<< "0x" << setw(2) << sig[2] << ") "
			<< "Signature from XML-file: (" << hex
			<< "0x" << setw(2) << sig0 << " "
			<< "0x" << setw(2) << sig1 << " "
			<< "0x" << setw(2) << sig2 << ").";

		throw new ErrorMsg( msg.str() );
	}

	return true;							// x閽߂B
}


bool AVRBootloader::writeFlashByte( long address, long value )
{
	setAddress( address >> 1 );				// ׯ͌ڽgp

	/* ڽȂްړ */
	if( address & 0x01 )					// ڽ?
		value = (value << 8) | 0x00ff;		// 1ׯ̏޲Ăړ
	else
		value |= 0xff00;					// ޲Ăɑ΂ďȂƂۏ

	/* ʂƏʂ޲Ă𑗏o */
	writeFlashLowByte( value & 0xff );
	writeFlashHighByte( value >> 8 );

	/* ߰ޏ݂𔭍s */
	setAddress( address >> 1 );				// ڽ͎ł܂B
	writeFlashPage();

	return true;							// x閽߂B
}


bool AVRBootloader::writeEEPROMByte( long address, long value )
{
	if( address >= 0x10000 )
		throw new ErrorMsg( "EEPROM addresses above 64k are currently not supported!" );

	setAddress( address );

	/* ްM */
	comm->sendByte( 'D' );
	comm->sendByte( value );
	comm->flushTX();

	/* CRԂׂ */
	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Writing byte to EEPROM failed! "
							"Programmer did not return CR after 'D'-command." );

	return true;							// x閽߂B
}


bool AVRBootloader::writeFlash( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	bool autoincrement;						// ް۰ނڽx?
	long address;

	/* ߰ޗeʂݒ肩 */
	if( pagesize == -1 )
		throw new ErrorMsg( "Programmer pagesize is not set!" );

	/* 򏑂ݎx */
	comm->sendByte( 'b' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
	{
		Util.log( "Using block mode...\r\n" );
		return writeFlashBlock( data );		// ݏI
	}

	/* HEX̧ق͈͎擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* x */
	comm->sendByte( 'a' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
		autoincrement = true;
	else
		autoincrement = false;

	/* ڽݒ */
	setAddress( start >> 1 );				// ׯ͌ڽgp

	/* 1̊޲ď݂Kv? */
	address = start;
	if( address & 1 )
	{
		/* ޲Ă݂̂gp */
		writeFlashLowByte( 0xff );			// ޲Ăŏ݂Ȃ
		writeFlashHighByte( data->getData( address ) );
		address++;

		/* ߰ޏ݂Kv? */
		if( (address % pagesize) == 0 ||
            address > end )					// x߰ދE܂͂Ə޲ĂȂ?
		{
			setAddress( (address-2) >> 1 );	// ߰ޓڽɐݒ
			writeFlashPage();
			setAddress( address >> 1 );
		}
	}

	/* ꏑ */
	while( (end-address+1) >= 2 )			// ƌc?
	{
		/* Ăѱڽݒ肪Kv? */
		if( !autoincrement )
			setAddress( address >> 1 );

		/* ꏑ */
		writeFlashLowByte( data->getData( address ) );
		writeFlashHighByte( data->getData( address+1 ) );
		address += 2;

		if( (address % MEM_PROGRESS_GRANULARITY) == 0 )
			Util.progress( "#" );			// i\is

		/* ߰ޏݕKv? */
		if( (address % pagesize) == 0 ||
            address > end )					// x߰ދE܂͂Ə޲ĂȂ?
		{
			setAddress( (address-2) >> 1 );	// ߰ޓڽɐݒ
			writeFlashPage();
			setAddress( address >> 1 );
		}
	}

	/* IO1̋޲ď݂Kv? */
	if( address == end )
	{
		/* ޲Ă̂ݎgp */
		writeFlashLowByte( data->getData( address ) );
		writeFlashHighByte( 0xff );			// ޲ĂŏȂ
		address+=2;

		/* ߰ޏ */
		setAddress( (address-2) >> 1 );		// ߰ޓڽɐݒ
		writeFlashPage();
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂B
}


bool AVRBootloader::writeFlashBlock( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	long blocksize;							// ް۰މ
	long bytecount;
	long address;

	/* P'b'ߔs'Y'ǂ܂ꂽƂ̉ŁAʂ擾 */
	blocksize = (comm->getByte() << 8) | comm->getByte();

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();
	
	/* 1̊޲ď݂Kv? */
	address = start;
	if( address & 1 )
	{
		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ޲Ă݂̂gp */
		writeFlashLowByte( 0xff );			// ޲Ăŏ݂Ȃ
		writeFlashHighByte( data->getData( address ) );
		address++;

		/* ߰ޏ݂Kv? */
		if( (address % pagesize) == 0 ||
			 address > end )				// x߰ދE܂͂Ə޲ĂȂ?
		{
			setAddress( (address-2) >> 1 );	// ߰ޓڽɐݒ
			writeFlashPage();
			setAddress( address >> 1 );
		}
	}

	/* ɉ̒Iւ݂̏Kv? */
	if( (address % blocksize) > 0 )			// ̒?
	{
		bytecount = blocksize - (address % blocksize);	// ޲Ă̎c

		if( (address+bytecount-1) > end )	// ݔ͈͂߂?
		{
			bytecount = end-address+1;		// ݔ͈͓޲Ă̎c
			bytecount &= ~0x01;				// ꐔ␳
		}

		if( bytecount > 0 )
		{
			setAddress( address >> 1 );		// ׯ͌ڽgp

			/* ׯ򏑂݊Jn */
			comm->sendByte( 'B' );
			comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
			comm->sendByte( bytecount & 0xff );
			comm->sendByte( 'F' );			// ׯ 

			while( bytecount > 0 )
			{
				comm->sendByte( data->getData( address ) );
				address++;
				bytecount--;
			}

			if( comm->getByte() != '\r' )
				throw new ErrorMsg( "Writing Flash block failed! "
									"Programmer did not return CR after 'BxxF'-command." );

			Util.progress( "#" );			// i\is
		}
	}

	/* ߂̊Sȉ򂪂ƗL? */
	while( (end-address+1) >= blocksize )
	{
		bytecount = blocksize;

		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ׯ򏑂݊Jn */
		comm->sendByte( 'B' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'F' );				// ׯ 

		while( bytecount > 0 )
		{
			comm->sendByte( data->getData( address ) );
			address++;
			bytecount--;
		}

		if( comm->getByte() != '\r' )
			throw new ErrorMsg( "Writing Flash block failed! "
								"Programmer did not return CR after 'BxxF'-command." );

		Util.progress( "#" );				// i\is
	}

	/* ŏIŉꂩ޲Ďc */
	if( (end-address+1) >= 1 )
	{
		bytecount = (end-address+1);		// ߂޲Ďc擾
		if( bytecount & 1 )
			bytecount++;					// ̌ꋫEɐ

		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ׯ򏑂݊Jn */
		comm->sendByte( 'B' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'F' );				// ׯ 

		while( bytecount > 0 )
		{
			if( address > end )
				comm->sendByte( 0xff );		// ݔ͈͊OɏĂ͂܂B
			else
				comm->sendByte( data->getData( address ) );

			address++;
			bytecount--;
		}

		if( comm->getByte() != '\r' )
			throw new ErrorMsg( "Writing Flash block failed! "
								"Programmer did not return CR after 'BxxF'-command." );

		Util.progress( "#" );				// i\is
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::readFlash( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	bool autoincrement;						// ް۰ނڽx?
	long address;

	if( pagesize == -1 )
		throw new ErrorMsg( "Programmer pagesize is not set!" );

	/* ǂݍݎx */
	comm->sendByte( 'b' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
	{
		Util.log( "Using block mode...\r\n" );
		return readFlashBlock( data );		// ݏI
	}

	/* HEX̧ق͈͎擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* x */
	comm->sendByte( 'a' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
		autoincrement = true;
	else
		autoincrement = false;

	/* ڽݒ */
	setAddress( start >> 1 );				// ׯ͌ڽgp

	/* 1̊޲ēǂݍ݂Kv? */
	address = start;
	if( address & 1 )
	{
		/* ǂݍ݁A޲Ă̂ݎgp */
		comm->sendByte( 'R' );
		comm->flushTX();

		data->setData( address, comm->getByte() );	// ޲
		comm->getByte();							// ޲ĕsgp
		address++;
	}

	/* 擾 */
	while( (end-address+1) >= 2 )
	{
		/* Ăѱڽݒ肪Kv? */
		if( !autoincrement )
			setAddress( address >> 1 );

		/* 擾 */
		comm->sendByte( 'R' );
		comm->flushTX();

		data->setData( address+1, comm->getByte() );	// ޲
		data->setData( address, comm->getByte() );		// ޲
		address += 2;

		if( (address % MEM_PROGRESS_GRANULARITY) == 0 )
			Util.progress( "#" );			// i\is

	};

	/* IO1̋޲ĂǂނƂKv? */
	if( address == end )
	{
		/* ǂݍ݁A޲Ă݂̂gp */
		comm->sendByte( 'R' );
		comm->flushTX();

		comm->getByte();							// ޲ĕsgp
		data->setData( address, comm->getByte() );	// ޲
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::readFlashBlock( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	long blocksize;							// ް۰މ
	long bytecount;
	long address;

	/* x'b'߂s'Y'ǂ܂ꂽƂ̉肵āAʎ擾 */
	blocksize = (comm->getByte() << 8) | comm->getByte();

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* ŏ1̊޲ĂǂނƂKv? */
	address = start;
	if( address & 1 )
	{
		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ʌꂾgp */
		comm->sendByte( 'R' );
		comm->flushTX();

		data->setData( address, comm->getByte() );	// ޲
		comm->getByte();							// ޲
		address++;
	}

	/* ŏɉ̒Ō֓ǂނƂKv? */
	if( (address % blocksize) > 0 )			// ̒?
	{
		bytecount = blocksize - (address % blocksize);	// ޲Ă̎c

		if( (address+bytecount-1) > end )	// ǂݍݔ͈͂߂?
		{
			bytecount = end-address+1;		// ǂݍݔ͈͓޲Ă̎c
			bytecount &= ~0x01;				// ꐔɕ␳
		}

		if( bytecount > 0 )
		{
			setAddress( address >> 1 );		// ׯ͌ڽgp

			/* ׯǂݍ݊Jn */
			comm->sendByte( 'g' );
			comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
			comm->sendByte( bytecount & 0xff );
			comm->sendByte( 'F' );			// ׯ 

			while( bytecount > 0 )
			{
				data->setData( address, comm->getByte() );
				address++;
				bytecount--;
			}

			Util.progress( "#" );			// i\is
		}
	}

	/* ǂނ߂̊Sȉ򂪂ƗL? */
	while( (end-address+1) >= blocksize )
	{
		bytecount = blocksize;

		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ׯǂݍ݊Jn */
		comm->sendByte( 'g' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'F' );				// ׯ 

		while( bytecount > 0 )
		{
			data->setData( address, comm->getByte() );
			address++;
			bytecount--;
		}

		Util.progress( "#" );				// i\is
	}

	/* ŏIɉ޲Ă̎c */
	if( (end-address+1) >= 1 )
	{
		bytecount = (end-address+1);		// ǂނ߂޲Ă̎c擾
		if( bytecount & 1 )
			bytecount++;					// ̌ꋫEɐ

		setAddress( address >> 1 );			// ׯ͌ڽgp

		/* ׯǂݍ݊Jn */
		comm->sendByte( 'g' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'F' );				// ׯ 

		while( bytecount > 0 )
		{
			if( address > end )
				comm->getByte();			// ݔ͈͊Oǂ܂Ȃ
			else
				data->setData( address, comm->getByte() );

			address++;
			bytecount--;
		}

		Util.progress( "#" );				// i\is
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::writeEEPROM( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	bool autoincrement;						// ް۰ނڽx?
	long address;

	/* 򏑂ݎx */
	comm->sendByte( 'b' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
	{
		Util.log( "Using block mode...\r\n" );
		return writeEEPROMBlock( data );	// ݏI
	}

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* x */
	comm->sendByte( 'a' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
		autoincrement = true;
	else
		autoincrement = false;

	/* ڽݒ */
	setAddress( start );

	/* ްM */
	address = start;
	do
	{
		/* Ăѱڽ̐ݒ肪Kv? */
		if( !autoincrement )
			setAddress( address );

		/* ޲đM */
		comm->sendByte( 'D' );
		comm->sendByte( data->getData( address ) );
		comm->flushTX();

		if( comm->getByte() != '\r' )
			throw new ErrorMsg( "Writing byte to EEPROM failed! "
								"Programmer did not return CR after 'D'-command." );

		if( (address % MEM_PROGRESS_GRANULARITY) == 0 )
			Util.progress( "#" );			// i\is

		address++;
	} while( address <= end );

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::writeEEPROMBlock( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	long blocksize;							// ް۰މ
	long bytecount;
	long address;

	/* x'b'߂s'Y'ǂ܂ꂽƉ肵āAʂ擾 */
	blocksize = (comm->getByte() << 8) | comm->getByte();

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* ްM */
	address = start;
	while( address <= end )					// ޲ĂƗL?
	{
		bytecount = blocksize;				// Sȉs

		if( (address+bytecount-1) > end )	// ݔ͈͂𒴂?
		{
			bytecount = end-address+1;		// ݔ͈͓޲Ă̎c
		}

		setAddress( address );

		/* EEPROM򏑂݊Jn */
		comm->sendByte( 'B' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'E' );				// EEPROM

		while( bytecount > 0 )
		{
			comm->sendByte( data->getData( address ) );
			comm->flushTX();

			address++;
			bytecount--;
		}

		if( comm->getByte() != '\r' )
			throw new ErrorMsg( "Writing EEPROM block failed! "
								"Programmer did not return CR after 'BxxE'-command." );

		Util.progress( "#" );				// i\is
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::readEEPROM( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	bool autoincrement;						// ް۰ނڽx?
	long address;

	/* 򏑂ݎx */
	comm->sendByte( 'b' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
	{
		Util.log( "Using block mode...\r\n" );
		return readEEPROMBlock( data );		// ݏI
	}

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* x */
	comm->sendByte( 'a' );
	comm->flushTX();

	if( comm->getByte() == 'Y' )
		autoincrement = true;
	else
		autoincrement = false;

	/* ڽݒ */
	setAddress( start );

	/* ްǂݍ */
	address = start;
	do
	{
		/* Ăѱڽݒ肪Kv? */
		if( !autoincrement )
			setAddress( address );

		/* ޲Ď擾 */
		comm->sendByte( 'd' );
		comm->flushTX();

		data->setData( address, comm->getByte() );

		if( (address % MEM_PROGRESS_GRANULARITY) == 0 )
			Util.progress( "#" );			// i\is

		address++;
	} while( address <= end );

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::readEEPROMBlock( HEXFile * data )
{
	long start, end;						// ް ڽ͈
	long blocksize;							// ް۰މ
	long bytecount;
	long address;

	/* x'b'߂s'Y'ǂ܂ꂽƉ肵āAʂ擾 */
	blocksize = (comm->getByte() << 8) | comm->getByte();

	/* HEX̧ق͈͂擾 */
	start = data->getRangeStart();
	end = data->getRangeEnd();

	/* ްǂݍ */
	address = start;
	while( address <= end )					// ǂ޲ĂƗL?
	{
		bytecount = blocksize;				// Sȉs

		if( (address+bytecount-1) > end )	// ǂޔ͈͂߂?
		{
			bytecount = end-address+1;		// ǂޔ͈͓޲Ă̎c
		}

		setAddress( address );

		/* EEPROMǂݍ݊Jn */
		comm->sendByte( 'g' );
		comm->sendByte( (bytecount>>8) & 0xff );	// 傫AMSBs
		comm->sendByte( bytecount & 0xff );
		comm->sendByte( 'E' );				// EEPROM

		while( bytecount > 0 )
		{
			data->setData( address, comm->getByte() );
			address++;
			bytecount--;
		}

		Util.progress( "#" );				// i\is
	}

	Util.progress( "\r\n" );				// i\I
	return true;							// x閽߂
}


bool AVRBootloader::writeLockBits( long bits )
{
	/* 'I'ߑM */
	comm->sendByte( 'l' );
	comm->sendByte( bits & 0xff );
	comm->flushTX();

	/* CRԂׂ */
	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Writing lock bits failed! "
							"Programmer did not return CR after 'l'-command." );

	return true;							// x閽߂
}


bool AVRBootloader::readLockBits( long * bits )
{
	/* 'r'߂𑗐M */
	comm->sendByte( 'r' );
	comm->flushTX();

	/* ް擾 */
	*bits = comm->getByte();

	return true;							// x閽߂
}


bool AVRBootloader::writeFuseBits( long bits )
{
	return false;							// xȂ߂
}


bool AVRBootloader::readFuseBits( long * bits )
{
	long lowfuse, highfuse;

	/* 'N'߂𑗐M */
	comm->sendByte( 'N' );
	comm->flushTX();

	/* ˭ޏ޲Ă擾 */
	highfuse = comm->getByte();

	/* 'F'߂𑗐M */
	comm->sendByte( 'F' );
	comm->flushTX();

	/* ˭މ޲Ă擾 */
	lowfuse = comm->getByte();

	*bits = (highfuse << 8) | lowfuse;

	return true;							// x閽߂
}


bool AVRBootloader::writeExtendedFuseBits( long bits )
{
	return false;							// xȂ߂
}


bool AVRBootloader::readExtendedFuseBits( long * bits )
{
	/* 'Q'߂𑗐M */
	comm->sendByte( 'Q' );
	comm->flushTX();

	/* ް擾 */
	*bits = comm->getByte();

	return true;							// x閽߂
}


bool AVRBootloader::programmerSoftwareVersion( long * major, long * minor )
{
	/* ĳŔԍ𓾂邽߂'V'߂𑗐M */
	comm->sendByte( 'V' );
	comm->flushTX();

	/* ް擾 */
	*major = comm->getByte();
	*minor = comm->getByte();

	return true;							// x閽߂
}


bool AVRBootloader::programmerHardwareVersion( long * major, long * minor )
{
	return false;							// xȂ߂
}


void AVRBootloader::setAddress( long address )
{
	/* ݂̱ڽݒ */
	if( address < 0x10000 ) {
		comm->sendByte( 'A' );
		comm->sendByte( (address >> 8) & 0xff );
		comm->sendByte( address & 0xff );
		comm->flushTX();
	} else {
		comm->sendByte( 'H' );
		comm->sendByte( (address >> 16) & 0xff );
		comm->sendByte( (address >> 8) & 0xff );
		comm->sendByte( address & 0xff );
		comm->flushTX();
	}

	/* CRԂׂ */
	if( comm->getByte() != '\r' ) {
		throw new ErrorMsg( "Setting address for programming operations failed! "
							"Programmer did not return CR after 'A'-command." );
	}
}


void AVRBootloader::writeFlashLowByte( long value )
{
	comm->sendByte( 'c' );
	comm->sendByte( value );
	comm->flushTX();

	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Writing Flash low byte failed! "
							"Programmer did not return CR after 'c'-command." );
}


void AVRBootloader::writeFlashHighByte( long value )
{
	comm->sendByte( 'C' );
	comm->sendByte( value );
	comm->flushTX();

	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Writing Flash high byte failed! "
							"Programmer did not return CR after 'C'-command." );
}


void AVRBootloader::writeFlashPage()
{
	comm->sendByte( 'm' );
	comm->flushTX();

	if( comm->getByte() != '\r' )
		throw new ErrorMsg( "Writing Flash page failed! "
							"Programmer did not return CR after 'm'-command." );
}


/* ̧ق̍Ō */
