/*****************************************************************************
 *
 * Atmel Corporation
 *
 * ̧		: SerialPort.cpp
 * ߲	: Dev-C++ 4.9.8.0 - http://bloodshed.net/dev/
 * 		: $Revision: 1163 $
 * t		: $Date: 2006N82 15:38:16 +0200 $
 * XV	: $Author: ohlia $
 *
 * xҰ	: avr@atmel.com
 *
 * ړI	: Win32
 *
 * pLq	: AVR911 - AVRJ݊
 *
 * 		: PCCOM߰ĂʂļرْʐM񋟂׽B̸׽CommChannel
 *			  ۸׽h܂B
 *
 ****************************************************************************/
#include "SerialPort.hpp"


/* ݽ׸ */
SerialPort::SerialPort( long _portNumber, long _timeout )
{
	if( _timeout < 0 )
		throw new ErrorMsg( "Negative COM-port timeout not allowed!" );

	if( _portNumber < 1 || _portNumber > 8 )
		throw new ErrorMsg( "Only COM1 to COM8 is supported!" );

	/* Ұ */
	portNumber = _portNumber;
	timeout = _timeout;
	channelOpen = false;
}


/* ޽׸ */
SerialPort::~SerialPort()
{
	closeChannel();
}


/* ʐMقJ */
void SerialPort::openChannel()
{
	char comName[] = "COMx";
	COMMTIMEOUTS comTimeouts;

	/* قɊJĂ邩 */
	if( channelOpen )
		throw new ErrorMsg( "Channel already open! Cannot open port twice." );

	/* COM̧ٖ𐶐ĊJƂs */
	comName[3] = '0' + portNumber;
	serialHandle = CreateFile( comName, GENERIC_READ | GENERIC_WRITE, 0, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	/* ߰ĂJƂsꍇɈُo͂ĕA */
	if( serialHandle == INVALID_HANDLE_VALUE )
		throw new ErrorMsg( "Error opening COM port!" );

	channelOpen = true;

	/* COM߰Đݒi[ */
	if( !GetCommTimeouts( serialHandle, &oldComTimeouts ) )
		throw new ErrorMsg( "Error reading COM port settings!" );

	/* COMݒ̕ʂ̕𓾂ĂύX */
	if( !GetCommTimeouts( serialHandle, &comTimeouts ) )
		throw new ErrorMsg( "Error reading COM port settings!" );

	comTimeouts.ReadIntervalTimeout = MAXDWORD;
	comTimeouts.ReadTotalTimeoutConstant = 0;
	comTimeouts.ReadTotalTimeoutMultiplier = 0;
	
	/* VݒKp */
	if( !SetCommTimeouts( serialHandle, &comTimeouts ) )
		throw new ErrorMsg( "Error changing COM port settings!" );
}


/* ʐMق */
void SerialPort::closeChannel()
{
	if( !channelOpen )
		return;

	/* COMҰ𕜌 */
	if( !SetCommTimeouts( serialHandle, &oldComTimeouts ) )
		throw new ErrorMsg( "Error changing COM port settings!" );

	/* ߰ĊJ */
	if( serialHandle != INVALID_HANDLE_VALUE )
		if( !CloseHandle( serialHandle ) )
			throw new ErrorMsg( "Error closing COM port!" );

	channelOpen = false;
}


/* P޲đM */
void SerialPort::sendByte( long data )
{
	DWORD written;

	/* قJĂ邩 */
	if( !channelOpen )
		throw new ErrorMsg( "Channel not open! Cannot send to unopened channel." );

	/* ݂s */
	if( !WriteFile( serialHandle, &data, 1, &written, NULL ) )
		throw new ErrorMsg( "Error writing byte to COM port!" );
}


/* P޲ĎM */
long SerialPort::getByte()
{
	time_t startTime;
	startTime = time( NULL );				// bŌݎԂǂ
	DWORD readnum;
	unsigned char data;

	/* قJĂ邩 */
	if( !channelOpen )
		throw new ErrorMsg( "Channel not open! Cannot read from unopened channel." );

	/* Ԑ߂܂޲ĎMs */
	do
	{
		/* ߰Ă޲ēǂݍ */
		if( !ReadFile( serialHandle, &data, 1, &readnum, NULL ) )
		{
			throw new ErrorMsg( "Error reading byte from COM port!" );
		}

		if( readnum == 1 )
			return ((long) data) & 0xff;

	} while( time(NULL) - startTime < timeout );

	/* Ԓ */
	throw new ErrorMsg( "Timeout during COM-port read operation!" );
}


/* MɏՕj */
void SerialPort::flushTX()
{
	/* قJĂ邩 */
	if( !channelOpen )
		throw new ErrorMsg( "Channel not open! Cannot flush an unopened channel." );

	/* ݊ɏՕް| */
	if( !PurgeComm( serialHandle, PURGE_TXCLEAR ) )
		throw new ErrorMsg( "Error flushing COM port TX buffer!" );
}


/* MɏՕj */
void SerialPort::flushRX()
{
	/* قJĂ邩 */
	if( !channelOpen )
		throw new ErrorMsg( "Channel not open! Cannot flush an unopened channel." );

	/* ǂݍ݊ɏՕް| */
	if( !PurgeComm( serialHandle, PURGE_RXCLEAR ) )
		throw new ErrorMsg( "Error flushing COM port RX buffer!" );
}


/* ޲đM */
void SerialPort::sendMultiple( unsigned char * data, long bufsize )
{
	DWORD written;

	/* قJĂ邩 */
	if( !channelOpen )
		throw new ErrorMsg( "Channel not open! Cannot write to unopened channel." );

	/* ݎs */
	if( !WriteFile( serialHandle, data, bufsize, &written, NULL ) )
		throw new ErrorMsg( "Error writing multiple bytes to COM port!" );
}

/* ̧ق̍Ō */
